#include "fw/servo/holder.h"
#include "emlabcpp/static_vector.h"

#pragma once

namespace em = emlabcpp;

namespace schpin
{

template < std::size_t N >
class servo_group
{

        em::static_vector< servo_holder, N > servos_;

        std::size_t poll_i_ = 0;

      public:
        bool any_has_task() const
        {
                return any_has_task( servo_flags{}.set() );
        }
        bool any_has_task( servo_flags filter ) const
        {
                return em::any_of( servos_, [&]( const auto& s ) {
                        return filter[s.get_id()] && s.has_task();
                } );
        }

        template < typename Visitor >
        auto make_matcher( Visitor&& vis )
        {
                return match(
                    [&]( control_tag, lewan_servo_id id, servo_master_message m ) {
                            vis( control_tag{}, fw_node_group::create< SERVO >( *id, m ) );
                    },
                    [&]( servo_tag, s_ctl_id msg_id, reply r, master_lewan_message m ) {
                            vis( servo_tag{}, msg_id, r, m );
                    },
                    [&]( activate_tag, servo_id id ) {
                            vis( control_tag{},
                                 fw_node_group::create< ACTIVATE >( id, ACTIVATED ) );
                    } );
        }

        template < typename Visitor >
        void on_servo_receive( clk_time                    clock_time,
                               s_ctl_id                    s_id,
                               const lewan_master_message& m,
                               Visitor&&                   vis )
        {
                lewan_master_protocol::extract( m )
                    .bind_left( em::uncurry |
                                [&]( lewan_servo_id id, lewan_master_submessage m ) {
                                        return lewan_master_group::extract( m ).convert_left(
                                            [&]( auto var ) {
                                                    return std::make_tuple( id, var );
                                            } );
                                } )
                    .match(
                        em::uncurry |
                            [&]( lewan_servo_id id, lewan_master_variant msg_var ) {
                                    auto iter = find_servo( *id );

                                    if ( iter == servos_.end() ) {
                                            reply_ctl_error< SERVO_MSG_UNKNOWN_E >(
                                                ctl_id{ 0 }, vis, m );
                                            return;
                                    }
                                    auto opt_msg =
                                        iter->on_servo_receive( clock_time, s_id, msg_var );
                                    if ( opt_msg ) {
                                            vis( control_tag{},
                                                 fw_node_group::create< SERVO >( *id,
                                                                                     *opt_msg ) );
                                    }
                            },
                        [&]( auto e ) {  // TODO: add error handling
                                SCHPIN_ERROR_LOG( "lewan.group", "Lewan error: " << e );
                        } );
        }

        template < typename Visitor >
        void on_servo_not_replying( const master_lewan_message& m, Visitor&& vis )
        {
                master_lewan_protocol::match(
                    m,
                    [&]( error_tag, auto... ) {  // TODO: add error handling

                    },
                    [&]( lewan_servo_id sid, master_lewan_submessage ) {
                            auto iter = find_servo( *sid );

                            if ( iter->is_activating() ) {
                                    using namespace std;
                                    swap( *iter, servos_.back() );
                                    servos_.pop_back();

                                    reply_ctl< ACTIVATE >( vis, *sid, MISSING );
                            }
                    } );
        }

        template < typename Visitor >
        void tick( clk_time clock_time, Visitor&& vis )
        {
                for ( auto& s : servos_ ) {
                        s.tick( clock_time, make_matcher( vis ) );
                }

                polling_tick( vis );
        }

        template < typename Visitor >
        void stop( Visitor&& v )
        {
                for ( auto& s : servos_ ) {
                        s.stop( v );
                }
        }

        template < typename Visitor >
        void activate( ctl_id cid, servo_addr sid, Visitor&& v )
        {
                if ( !std::holds_alternative< lewan_servo_id >( sid ) ) {
                        // TODO: EROOOOOOOOR
                        return;
                }
                lewan_servo_id id   = std::get< lewan_servo_id >( sid );
                auto           iter = find_servo( *id );
                if ( iter != servos_.end() ) {
                        if ( iter->is_activating() ) {
                                reply_ctl< ACTIVATE >( v, *id, WAS_ACTIVATING );
                        } else {
                                reply_ctl< ACTIVATE >( v, *id, WAS_ACTIVATED );
                        }
                        return;
                }

                if ( servos_.size() == servos_.max_size() ) {
                        reply_ctl_error< SERVO_CAPACITY_USED_E >( cid, v );
                        return;
                }

                servos_.emplace_back( id );

                reply_ctl< ACTIVATE >( v, *id, ACTIVATING );
        }

        template < typename Visitor >
        void on_event( ctl_id id, clk_time clock_time, servo_event e, Visitor&& vis )
        {
                auto iter = find_servo( e.id );
                // TODO: check!
                iter->on_event( id, clock_time, e.sub_event, make_matcher( vis ) );
        }

        em::either< servo_event, fw_node_group_message >
        on_servo_packet( ctl_id id, servo_id sid, master_servo_command_variant subvar )
        {
                auto iter = find_servo( sid );
                if ( iter == servos_.end() ) {
                        return { make_ctl_error< SERVO_NOT_ACTIVE_E >( id, sid ) };
                }

                return iter->on_servo_packet( id, subvar )
                    .convert_right( [&]( servo_master_message m ) {
                            return fw_node_group::create< SERVO >( sid, m );
                    } );
        }

      private:
        template < typename Visitor >
        void polling_tick( Visitor&& vis )
        {
                if ( servos_.empty() ) {
                        return;
                }
                poll_i_ = ( poll_i_ + 1 ) % servos_.size();
                servos_[poll_i_].poll( poll_i_, vis );
        }

        auto find_servo( servo_id id )
        {
                return em::find_if( servos_, [&]( const auto& j ) {
                        return j.get_id() == id;
                } );
        }
};

}  // namespace schpin
