#include "schpin_lib/logging.h"
//
#include "blob/reactor.h"
#include "test/servo.h"

#include <gtest/gtest.h>

using namespace schpin;

class ReactorFixture : public ::testing::Test
{
      public:
        blob_reactor< 20 >                                             reactor;
        prot_config_loader< blob_config >                              b_config;
        std::map< lewan_servo_id, sim_servo >                          servos;
        std::map< lewan_servo_id, prot_config_loader< servo_config > > configs;
        clk_time                                                       system_time;

        void SetUp() override
        {
                b_config->input_buffer_telem_period_ms = clk_time{ 2000 };
                b_config->event_buffer_telem_period_ms = clk_time{ 2000 };

                for ( std::size_t i : em::range( 3u ) ) {
                        servos.emplace( *lewan_servo_id::make( i ),
                                        sim_servo{ *lewan_servo_id::make( i ), 0.f } );

                        configs.emplace( *lewan_servo_id::make( i ),
                                         prot_config_loader< servo_config >(  //
                                             servo_config{ bangle::from_float( 0.f ),
                                                           bangle::from_float( 4.18879f ),
                                                           lewan_angle::get< 0 >(),
                                                           lewan_angle::get< 1000 >(),
                                                           lewan_angle::get< 5 >(),
                                                           lewan_angle::get< 10 >() } ) );
                }
        }

        auto make_matcher()
        {
                return match(
                    [this]( control_tag, blob_master_group_message group_msg ) -> void {
                            on_master_msg( group_msg );
                    },
                    [&]( servo_tag, reply, const master_lewan_message& m ) -> void {
                            pass_to_sim( m );
                    } );
        }

        void tick( std::size_t n )
        {
                for ( std::size_t i = 0; i < n; ++i ) {
                        system_time += clk_time{ 1 };
                        reactor.tick( system_time, make_matcher() );
                }
        }

        void on_master_msg( blob_master_group_message m )
        {
                blob_master_group::extract( m ).match(
                    [&]( blob_master_group_variant v ) {
                            v << tuple_match(
                                [&]( cmd_tag< SERVO >,
                                     servo_id                     id,
                                     servo_master_command_variant var ) {
                                        on_servo_ctl_message( id, var );
                                },
                                [&]( cmd_tag< CONFIG >, blob_master_config_variant var ) {
                                        b_config.on_servo_config_message(
                                            var, [&]( master_blob_config_message m ) {
                                                    ctl_to_reactor(
                                                        master_blob_group::create< CONFIG >( m ) );
                                            } );
                                },
                                [&]( cmd_tag< ACTIVATE >, servo_id, activated_enum ) {},
                                [&]( cmd_tag< EVENT_BUFFER_SIZE >, uint16_t ) {},
                                [&]( cmd_tag< INPUT_BUFFER_SIZE >, uint16_t ) {},
                                [&]( cmd_tag< FINISHED >, ctl_id ) {},
                                [&]( auto... ) { FAIL() << m; } );
                    },
                    [&]( auto... ) { FAIL(); } );
        }

        void on_servo_ctl_message( servo_id id, servo_master_command_variant var )
        {
                // TODO: copypasted from lewan servo gtest
                var << tuple_match(
                    [&]( cmd_tag< CONFIG >, servo_master_config_variant cvar ) {
                            configs.at( *lewan_servo_id::make( id ) )
                                .on_servo_config_message(
                                    cvar, [&]( master_servo_config_message msg ) {
                                            on_config_message( *lewan_servo_id::make( id ), msg );
                                    } );
                    },
                    [&]( cmd_tag< POSITION >, bangle ) {},
                    [&]( auto cmd, auto... ) { FAIL() << "with arg: " << cmd; } );
        }

        void on_config_message( lewan_servo_id id, master_servo_config_message msg )
        {
                auto full_servo_msg = master_servo_group::create< CONFIG >( msg );
                auto full_msg =
                    master_blob_group::create< SERVO >( servo_id{ *id }, full_servo_msg );
                ctl_to_reactor( full_msg );
        }

        void ctl_to_reactor( master_blob_group_message full_msg )
        {
                master_blob_group::extract( full_msg )
                    .match(
                        [&]( master_blob_group_variant cmd ) {
                                reactor.on_control_receive( ctl_id{ 0 }, cmd, make_matcher() );
                        },
                        [&]( auto... ) { FAIL(); } );
        }

        void pass_to_sim( const master_lewan_message& m )
        {
                // TODO: copypates from lewan_servo_getst
                master_lewan_protocol::extract( m )
                    .bind_left( em::uncurry |
                                [&]( lewan_servo_id id, auto sub_msg ) {  //
                                        return master_lewan_group::extract( sub_msg ).convert_left(
                                            [&]( auto var ) {  //
                                                    return std::make_tuple( id, var );
                                            } );
                                } )
                    .match(
                        em::uncurry |
                            [&]( lewan_servo_id id, master_lewan_variant v ) {
                                    auto opt_msg = servos.at( id ).on_cmd( system_time, id, v );
                                    if ( !opt_msg ) {
                                            return;
                                    }
                                    reactor.on_servo_receive(
                                        system_time, *opt_msg, make_matcher() );
                            },
                        [&]( protocol_error_record ) {  //
                                FAIL();
                        } );
        }

        template < auto ID, typename... Args >
        void send_to_react( Args... args )
        {
                auto msg = master_blob_group::create< ID >( args... );
                // TODO: m,ake protocol method for creation of variant
                master_blob_group::extract( msg ).match(
                    [&]( auto var ) {  //
                            reactor.on_control_receive( ctl_id{ 0 }, var, make_matcher() );
                    },
                    [&]( auto ) {  //
                            FAIL();
                    } );
        }

        template < auto ID, typename... Args >
        void send_servo_to_react( servo_id id, Args... args )
        {
                send_to_react< SERVO >( id, master_servo_group::create< ID >( args... ) );
        }
};

TEST_F( ReactorFixture, basic )
{
        tick( 150 );
        for ( std::size_t i : em::range( 3u ) ) {
                send_to_react< ACTIVATE >( static_cast< servo_id >( i ) );
                tick( 30 );
        }
}
TEST_F( ReactorFixture, syncing )
{
        tick( 30 );
        for ( std::size_t i : em::range( 3u ) ) {
                send_to_react< ACTIVATE >( static_cast< servo_id >( i ) );
                tick( 30 );
        }

        send_servo_to_react< GOTO >( 1, bangle::from_float( 1.f ), btime::from_float( 1.f ) );
        tick( 5 );
        servo_flags b1;
        b1[1] = true;
        b1[2] = true;
        send_to_react< SYNC >( b1 );
        send_servo_to_react< GOTO >( 0, bangle::from_float( 1.f ), btime::from_float( 1.f ) );
        tick( 5 );
        send_servo_to_react< GOTO >( 1, bangle::from_float( 2.f ), btime::from_float( 1.f ) );
        tick( 5 );
        send_servo_to_react< GOTO >( 2, bangle::from_float( 2.f ), btime::from_float( 1.f ) );
        tick( 5 );
        servo_flags b2;
        b2[0] = true;
        b2[1] = true;
        b2[2] = true;
        send_to_react< SYNC >( b2 );
        send_servo_to_react< GOTO >( 0, bangle::from_float( 3.f ), btime::from_float( 1.f ) );
        tick( 5 );

        // 25 ms ellapse at this  point
        tick( 500 );

        lewan_servo_id S0 = lewan_servo_id::get< 0 >();
        lewan_servo_id S1 = lewan_servo_id::get< 1 >();
        lewan_servo_id S2 = lewan_servo_id::get< 2 >();

        // 0/1 are moving , because barrier b1 only blocks servo 2/1, not 0
        EXPECT_NE( servos.at( S0 ).get_rad_angle(), 0.f );
        EXPECT_NE( servos.at( S1 ).get_rad_angle(), 0.f );
        EXPECT_EQ( servos.at( S2 ).get_rad_angle(), 0.f );

        tick( 500 );
        // now 1025ms has passed

        EXPECT_LT( std::fabs(servos.at( S0 ).get_rad_angle() - 1.f ), 0.01f);
        EXPECT_NE( servos.at( S1 ).get_rad_angle(), 0.f );
        EXPECT_NE( servos.at( S2 ).get_rad_angle(), 0.f );

        tick( 1000 );
        // now 2025ms has passed

        EXPECT_LT( std::fabs(servos.at( S0 ).get_rad_angle() - 1.f ), 0.01f);
        EXPECT_LT( std::fabs(servos.at( S1 ).get_rad_angle() - 2.f ), 0.01f);
        EXPECT_LT( std::fabs(servos.at( S2 ).get_rad_angle() - 2.f ), 0.2f);

	tick( 1100 );
	// now 3125ms has passed
        EXPECT_LT( std::fabs(servos.at( S0 ).get_rad_angle() - 3.f ), 0.3f);
        EXPECT_LT( std::fabs(servos.at( S1 ).get_rad_angle() - 2.f ), 0.01f);
        EXPECT_LT( std::fabs(servos.at( S2 ).get_rad_angle() - 2.f ), 0.01f);
}

int main( int argc_in, char** argv_in )
{
        testing::InitGoogleTest( &argc_in, argv_in );

        return RUN_ALL_TESTS();
}
